#include<iostream>

#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/Module.h"
#include "llvm/LinkAllPasses.h"
#include "llvm/Analysis/CFGPrinter.h"
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/Pass.h"

#include <clang/CodeGen/CodeGenAction.h>
#include <clang/Frontend/CompilerInstance.h>
#include <clang/Frontend/CompilerInvocation.h>
#include <clang/Basic/DiagnosticOptions.h>
#include <clang/Basic/TargetInfo.h>
#include <vector>

std::unique_ptr<llvm::Module> c2ir( char * filename,
                                    llvm::LLVMContext& llvm_ctx ) {

  std::vector<const char *> args;
  args.push_back( "-emit-llvm" );
  args.push_back( "-disable-llvm-passes" );
  args.push_back( "-debug-info-kind=standalone" );
  args.push_back( "-dwarf-version=2" );
  args.push_back( "-dwarf-column-info" );
  args.push_back( "-mdisable-fp-elim");
  args.push_back( "-femit-all-decls" );
  args.push_back( "-O1" );
  args.push_back( "-disable-O0-optnone" );
  args.push_back( filename );

  clang::CompilerInstance Clang;
  Clang.createDiagnostics();

  std::shared_ptr<clang::CompilerInvocation> CI(new clang::CompilerInvocation());
  clang::CompilerInvocation::CreateFromArgs( *CI.get(), &args[0],
                                             &args[0] + args.size(),
                                            Clang.getDiagnostics());
  Clang.setInvocation(CI);
  clang::CodeGenAction *Act = new clang::EmitLLVMOnlyAction(&llvm_ctx);

  if (!Clang.ExecuteAction(*Act))
    return nullptr;

  std::unique_ptr<llvm::Module> module = Act->takeModule();

  return std::move(module);
}


void prepare( std::unique_ptr<llvm::Module>& m ) {

  llvm::legacy::PassManager pm;
  pm.add( llvm::createPromoteMemoryToRegisterPass() );
  pm.add( llvm::createCFGPrinterLegacyPassPass() );
  pm.run( *m.get() );

}


class my_first_pass : public llvm::FunctionPass {
  public:
  static char ID;

  my_first_pass() : llvm::FunctionPass(ID) {}

  bool runOnFunction(llvm::Function &f) {
    std::string name = f.getName();
    std::cout << "Processing function : " << name << "\n";
    // for( const llvm::BasicBlock& b : f ) {
    //   for( const llvm::Instruction& Iobj : b.getInstList() ) {
    //     const llvm::Instruction* I = &(Iobj);
    //     I->print( llvm::outs() );
    //     // if(auto bop = llvm::dyn_cast<llvm::BinaryOperator>(I) ) {
    //     //   auto op0 = bop->getOperand( 0 );
    //     //   auto op1 = bop->getOperand( 1 );
    //     //   unsigned op = bop->getOpcode();
    //     //   if( op == llvm::Instruction::Add ) {
    //     //     std::cout << "Add instruction found: \n";
    //     //     bop->print( llvm::outs() );
    //     //     std::cout << "\ninput 0 : \n";
    //     //     op0->print( llvm::outs() );
    //     //     std::cout << "\ninput 1 : \n";
    //     //     op1->print( llvm::outs() );
    //     //     std::cout << "\n";
    //     //   }
    //     // }
    //   }
    // }
    return false;
  }

  void getAnalysisUsage(llvm::AnalysisUsage &au) const {
    au.setPreservesAll();
    // au.addRequired<llvm::LoopInfoWrapperPass>();
  }

  llvm::StringRef getPassName() const {
    return "my first pass, let us see what can we do with it!";
  }
};
char my_first_pass::ID = 0;


void run_my_pass( std::unique_ptr<llvm::Module>& m ) {
  llvm::legacy::PassManager pm;
  pm.add( new my_first_pass() );
  pm.run( *m.get() );
}

int main(int argc, char** argv) {
  if( argc != 2 ) {
    return 1;
  }
  char* cpp_file = argv[1];

  llvm::LLVMContext llvm_ctx;

  std::unique_ptr<llvm::Module> m = c2ir( cpp_file, llvm_ctx );

  // prepare( m );

  // run_my_pass( m );

  // m->print( llvm::outs(), nullptr );
}
